import math
import fractions
import decimal
from collections import namedtuple
import forallpeople.tuplevector as vec
import pytest

Point = namedtuple("Point", ["x", "y", "z"])
Dimensions = namedtuple("Dimensions", ["kg", "m", "s", "A", "cd", "K", "mol"])

P1 = Point(2.0, 3.4, 1.0)
P2 = Point(1.5, 2.8, 4.5)
P3 = Point(1, 1, 0)
P4 = Point(1, 0, 0)
P5 = Point(-1, 0, 0)
P6 = Point(3, 4, 0)
P7 = Point(0,0,0)
D1 = Dimensions(1., 1., -2., 0,0,0,0)
D2 = Dimensions(2., 2., -4., 0,0,0,0)
D3 = Dimensions(1., 2., -2., 0,0,0,0)
D4 = Dimensions(-1, 1, 0, 0, 0, 0, 0)
T1 = (4.5, 1.1112, 2)
T2 = (1.2, 4, 3.267)
T3 = (3, 4, 0)
T4 = (0,0,0)
T5 = (1/2, math.sqrt(3)/2, 0)
T6 = (1/2, 0, 0)
NT1 = ("A", 2, 3.4)
NT2 = ("B", "C", "D")
L1 = tuple(range(0,100))
L2 = tuple(range(1,101))
LT1 = list(range(0,100))
LT2 = list(range(1, 101))

def test_valid_for_arithmetic():
    assert vec.valid_for_arithmetic(134) == True
    assert vec.valid_for_arithmetic(23.5) == True
    assert vec.valid_for_arithmetic(complex(3,4)) == True
    assert vec.valid_for_arithmetic("apple") == False
    assert vec.valid_for_arithmetic(fractions.Fraction(3,4)) == True
    assert vec.valid_for_arithmetic(decimal.Decimal(3.123145645879)) == True
    assert vec.valid_for_arithmetic(P1) == True
    assert vec.valid_for_arithmetic(T1) == True
    assert vec.valid_for_arithmetic(LT1) == True
    
def test_tuple_check():
    with pytest.raises(ValueError):
        vec.tuple_check(3, T1)
        vec.tuple_check(LT1, 3)
        vec.tuple_check(NT1)
        vec.tuple_check(D1, NT2)
        vec.tuple_check(T3, NT2)
        vec.tuple_check(T3, "A")
        vec.tuple_check((1,), (2,), (3,))
    
def test_same_shape():
    assert vec.same_shape(P1, D1) == False
    assert vec.same_shape(T2, L1) == False
    assert vec.same_shape(NT1, NT2) == True
    assert vec.same_shape(T1, P3) == True
    
def test_dot():
    assert vec.dot(P1, P3) == 5.4
    assert vec.dot(D1, D3) == 7.0
    assert vec.dot(T1, T2) == 16.3788
    
def test_cross():
    assert vec.cross(P1, P3) == Point(x=-1.0, y=1.0, z=-1.4)
    assert vec.cross(T1, T2) == (-4.3697096, -12.301499999999999, 16.66656)
    
def test_add():
    assert vec.add(P1, P3) == Point(x=3.0, y=4.4, z=1.0)
    assert vec.add(D2, D3) == Dimensions(kg=3.0, m=4.0, s=-6.0, A=0.0, cd=0.0, K=0.0, mol=0.0)
    assert vec.add(L1, L2) == tuple(range(1, 200, 2))
    
def test_subtract():
    assert vec.subtract(P1, P3) == Point(x=1.0, y=2.4, z=1.0)
    assert vec.subtract(D2, D3) == Dimensions(kg=1.0, m=0.0, s=-2.0, A=0.0, cd=0.0, K=0.0, mol=0.0)
    assert vec.subtract(L1, L2) == tuple(-1 for n in range(0,100))
    
def test_multiply():
    assert vec.multiply(P1, P3) == Point(x=2.0, y=3.4, z=0.0)
    assert vec.multiply(D2, D3) == Dimensions(kg=2.0, m=4.0, s=8.0, A=0.0, cd=0.0, K=0.0, mol=0.0)
    assert vec.multiply(P3, 3.5) == Point(x=3.5, y=3.5, z=0)
    
def test_divide():
    assert str(vec.divide(P1, P3)) == str(Point(x=2.0, y=3.4, z=float("nan")))
    assert vec.divide(D2, D1, ignore_zeros = True) == Dimensions(*(2, 2, 2, 0,0,0,0))
    assert vec.divide(D2, 2) == D1
    
def test_magnitude():
    assert vec.magnitude(P6) == 5.0
    assert vec.magnitude(P7) == 0
    assert vec.magnitude(T3) == 5.0
    assert vec.magnitude(P3) == math.sqrt(2)
    assert vec.magnitude(D4) == math.sqrt(2)
    
def test_vround():
    assert vec.vround(T1, 3) == (4.5, 1.111, 2)
    assert vec.vround(P2, 0) == Point(x=2, y=3, z=4)
    
def test_mean():
    assert vec.mean(P6) == 2.3333333333333335
    assert vec.mean(D1) == 0
    assert vec.mean(T2) == 2.8223333333333334
    assert vec.mean(P6, True) == 3.5
    assert vec.mean(L1) == 49.5
    
def test__clip():
    assert vec._clip(-1) == -1
    assert vec._clip(-1.00001) == -1
    assert vec._clip(1.5) == 1
    assert vec._clip(-0.3) == -0.3
    assert vec._clip(0.4) == 0.4

def test_normalize():
    assert vec.normalize(P1) == Point(x=0.49147318718299043, y=0.8355044182110838, z=0.24573659359149522)
    assert vec.normalize(T1) == (0.8913991044053705, 0.22011615218116617, 0.39617737973572026)
    assert vec.normalize(D3) == Dimensions(kg=0.3333333333333333, m=0.6666666666666666, s=-0.6666666666666666, A=0.0, cd=0.0, K=0.0, mol=0.0)
    
def test_angle():
    assert vec.angle(P3, P4) == 0.7853981633974484
    assert vec.angle(P3, P4, True) == 45.00000000000001
    assert vec.angle(T5, T6) == 1.0471975511965976
    assert vec.angle(T5, T6, True) == 59.99999999999999


"""
tuplevector: Treat tuples of any kind (e.g. namedtuple, NamedTuple) 
like one dimensional vectors!
by Connor Ferster 03/2019
"""

# TODO: Explore idea of removing all checks entirely and using a decorator
# function to wrap all vector functions (with try/excepts) to capture errors 
# such as tuples of different lengths, elements that do not do math, or 
# other type errors. Could vastly speed performance.

from math import pi, acos, sqrt
from numbers import Number
from typing import Union, Any, Optional

def same_shape(t1: tuple, t2: tuple) -> bool:
    """
    Returns True if t1 and t2 are the same shape. 
    False, otherwise."""
    try:
        return len(t1) == len(t2)
    except:
        return False

def valid_for_arithmetic(other: Any) -> bool:
    """
    Returns True if object 'other' is valid for arithmetic operations. 
    Returns False otherwise.
    """
    try:
        sum(other) # This not a strong check but efficient for collections
        return True
    except TypeError:
        if isinstance(other, Number):
            return True
        return False

def tuple_check(t1: tuple, other: Any = None) -> Optional[bool]:
    """
    Returns None. Raises error if any of the tuple validation tests fail.
    """
    is_t1_tuple = isinstance(t1, tuple)
    is_other_tuple = isinstance(other, tuple)
    is_t1_valid = valid_for_arithmetic(t1)
    is_other_valid = valid_for_arithmetic(other)
    if (is_t1_tuple and is_t1_valid and other is None):
        return True
    if (is_t1_tuple and is_t1_valid and is_other_valid and is_other_tuple):
        return (True, True)
    elif (is_t1_tuple and is_t1_valid and is_other_valid and not is_other_tuple):
        return (True, False)
    elif not is_t1_tuple: 
        raise ValueError(f"Input object, {t1}, is not a valid tuple: first arg must be tuple.")
    elif not is_t1_valid:
        raise ValueError(f"Input tuple {t1} is not valid for arithmetic operations.")
    elif other is not None:
        if is_other_tuple and not same_shape(t1, other):
            raise ValueError(f"Input tuples must be same shape, not {len(t1)} and {len(other)}.")
        if is_other_tuple and not is_other_valid:
            raise ValueError(f"Input tuple, {other}, is not valid for arithmetic operations.")
        if not is_other_tuple and not valid_for_arithmetic(other):
            raise ValueError(f"Input object, {other}, is not valid for arithmetic operations.")

def dot(t1: tuple, t2: tuple) -> float:
    """
    Returns the dot product of the tuples, 't1' and 't2'.
    """
    _, t2_tup = tuple_check(t1, t2)
    if t2_tup:
        return sum([elem * t2[idx] for idx, elem in enumerate(t1)])
    else: 
        raise ValueError(f"Input tuples must be the same length. Got: {t1}, {t2}")
    
def cross(t1: tuple, t2: tuple) -> tuple:
    """
    Returns the cross product of two 3-dimensional tuples, 't1' and 't2'.
    (Raises error if tuples are not 3-dimensional). Maintains the tuple type
    't1' (e.g. if it is a namedtuple).
    """
    t1_tup, other_tup = tuple_check(t1, t2)
    if len(t1) == len(t2) == 3:
        i = t1[1]*t2[2] - t2[1]*t1[2]
        j = -(t1[0]*t2[2] - t2[0]*t1[2])
        k = t1[0]*t2[1] - t2[0]*t1[1]
        result = (i, j, k)
        try:
            return type(t1)(*result)
        except TypeError:
            return type(t1)(result)
    else:
        raise TypeError("Input tuples must be 3-dimensional. Got: {t1}, {t2}".format(t1=t1, t2=t2))
        
def add(t1: tuple, other: Union[tuple, int, float]) -> tuple:
    """
    Returns a tuple of element-wise multiplication of 't1' and 'other'
    """
    _, other_tup = tuple_check(t1, other)
    if other_tup:
        result = tuple(elem + other[idx] for idx, elem in enumerate(t1))
        try: 
            return type(t1)(*result)
        except TypeError:
            return type(t1)(result)
    else:
        result = (elem + other for elem in t1)
        try:
            print("3")
            return type(t1)(*result)
        except TypeError:
            print("4")
            return type(t1)(result)


def subtract(t1: tuple, other: Union[tuple, int, float]) -> tuple:
    """
    Returns a tuple of element-wise multiplication of 't1' and other
    """
    _, other_tup = tuple_check(t1, other)
    if other_tup:
        result = tuple(elem - other[idx] for idx, elem in enumerate(t1))
        try: 
            return type(t1)(*result)
        except TypeError:
            return type(t1)(result)
    else:
        result = (elem - other for elem in t1)
        try:
            return type(t1)(*result)
        except TypeError:
            return type(t1)(result)
        
def multiply(t1: tuple, other: Union[tuple, int, float]) -> tuple:
    """
    Returns a tuple of element-wise multiplication of 't1' and other
    """
    _, other_tup = tuple_check(t1, other)
    if other_tup:
        result = tuple(elem * other[idx] for idx, elem in enumerate(t1))
        try: 
            return type(t1)(*result)
        except TypeError:
            return type(t1)(result)
    else:
        result = (elem * other for elem in t1)
        try:
            return type(t1)(*result)
        except TypeError:
            return type(t1)(result)
    
def divide(t1: tuple, other: Union[tuple, int, float], 
           ignore_zeros:bool = False) -> tuple:    
    """
    Returns a tuple of element-wise division of 't1' and 't2'.
    If 'ignore_empty' is set to False, then the division will
    ignore elements of 0/0 and will return zero for those
    elements instead.
    """
    t1_tup, other_tup = tuple_check(t1, other)
    if other_tup:
        if ignore_zeros:
            result = tuple(
                elem / other[idx]
                if other[idx] != 0
                else 0 
                for idx, elem in enumerate(t1)
                )
        else:
            result = tuple(
                elem / other[idx]
                if other[idx] != 0
                else float('nan') 
                for idx, elem in enumerate(t1)
                )
        try:
            return type(t1)(*result)
        except TypeError:
            return type(t1)(result)
    else:
        if ignore_zeros:
            result = tuple(
                elem / other
                if other != 0
                else 0 
                for idx, elem in enumerate(t1)
            )
        else:
            result = tuple(
                elem / other
                if other != 0 
                else 0 
                for idx, elem in enumerate(t1)
            )
        try:
            return type(t1)(*result)
        except TypeError:
            return type(t1)(result)


    # acc = {}
    # if valid_for_arithmetic(other) and not isinstance(other, tuple):
    #     tuple_check(t1)
    #     for idx, val in enumerate(t1):
    #         if val == 0 and other == 0:
    #             acc.update({idx: float("nan")})
    #         elif other == 0:
    #             acc.update({idx: float("inf")})
    #         else:
    #             acc.update({idx: val/other})
    # else: 
    #     tuple_check(t1, other)
    #     for idx, val in enumerate(t1):
    #         if val == 0 and other[idx] == 0:
    #             if ignore_zeros:
    #                 acc.update({idx: 0})
    #             else:
    #                 acc.update({idx: float("nan")})
    #         elif other[idx] == 0:
    #             acc.update({idx: float("inf")})
    #         else:
    #             acc.update({idx: val/other[idx]})
    # return collapse_to_tuple(acc, type(t1))

def vround(t: tuple, precision: int = 0) -> tuple:
    """
    Returns a tuple with elements rounded to 'precision'.
    """
    t_tup = tuple_check(t)
    if t_tup:
        result = tuple(round(elem, precision) for elem in t)
        try:
            return type(t)(*result)
        except TypeError:
            return type(t)(result)


def mean(t: tuple, ignore_empty: bool = False) -> float:
    """
    Returns the average of the values in the tuple, 't'. If 'ignore_empty' 
    is True, then only the values that are not either 0 or None are averaged.
    If 'ignore_empty' is False, all values are used and None is taken as 0. 
    """
    tuple_check(t)
    count = 0
    total = 0
    for val in t:
        if ignore_empty: 
            if val == 0 or val == None:
                continue
            else:
                total += val
                count += 1
        else:
            total += val
            count += 1  
    count = count or 1 # in case t is an empty vector
    return total/count

def magnitude(t: tuple) -> float:
    """
    Returns the vector magnitude of the tuple, 't' using the 
    Pythagorean method.
    """
    tuple_check(t)
    mag_sqr = 0
    for val in t:
        mag_sqr += val**2
    return sqrt(mag_sqr)
    
def normalize(t: tuple) -> tuple:
    """
    Returns the normalized unit vector of the vector tuple, 't'.
    """
    return divide(t, magnitude(t))

def _clip(n: float) -> float:
    """
    Helper function to emulate numpy.clip for the specific
    use case of preventing math domain errors on the 
    acos function by "clipping" values that are > abs(1).
    e.g. _clip(1.001) == 1
         _clip(-1.5) == -1
         _clip(0.80) == 0.80
    """
    sign = n / abs(n)
    if abs(n) > 1: return 1 * sign
    else: return n
        

def angle(t1: tuple, t2: tuple, degrees: bool = False) -> float:
    """
    Returns the angle between two vector tuples, 't1' and 't2',
    in rads. If 'degrees' = True, then returned in degrees.
    """
    tuple_check(t1,t2)
    rad2deg = 1.0
    if degrees: rad2deg = 180/pi
    denom = (magnitude(t1) * magnitude(t2))
    if denom == 0: return pi/4 * rad2deg
    else: return acos(_clip(dot(t1, t2) / (magnitude(t1) * magnitude(t2)))) * rad2deg
    


#    Copyright 2020 Connor Ferster

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
The SI Units: "For all people, for all time"

A module to model the seven SI base units:

                    kg

            cd               m


                    SI
         mol                    s



               K           A

  ...and other derived and non-SI units for practical calculations.
"""

__version__ = "2.3.0"

from typing import Union, Optional


from forallpeople.dimensions import Dimensions
import forallpeople.physical_helper_functions as phf
import forallpeople.tuplevector as vec
from forallpeople.environment import Environment
import builtins
import sys
import warnings
NUMBER = (int, float)


class Physical(object):
    """
    A class that defines any physical quantity that can be described
    within the BIPM SI unit system.
    """

    _eps = 1e-7
    _total_precision = 6

    __slots__ = ("value", "dimensions", "factor", "precision", "prefixed")

    def __init__(
        self,
        value: Union[int, float],
        dimensions: Dimensions,
        factor: float,
        precision: int = 3,
        prefixed: str = "",
    ):
        """Constructor"""
        super(Physical, self).__setattr__("value", value)
        super(Physical, self).__setattr__("dimensions", dimensions)
        super(Physical, self).__setattr__("factor", factor)
        super(Physical, self).__setattr__("precision", precision)
        super(Physical, self).__setattr__("prefixed", prefixed)
    def __setattr__(self, _, __):
        raise AttributeError("Cannot set attribute.")

    ### API Methods ###
    @property
    def latex(self) -> str:
        return self._repr_latex_()

    @property
    def html(self) -> str:
        return self._repr_html_()

    def prefix(self, prefixed: str = ""):
        """
        Return a Physical instance with 'prefixed' property set to 'prefix'
        if 'prefixed' is set to "unity" then the unit will be forced into its
        unprefixed state.
        """
        if self.factor != 1:
            raise AttributeError("Cannot set a prefix on a Physical if it has a factor.")
        # check if elligible for prefixing; do not rely on __repr__ to ignore it
        return Physical(
            self.value, self.dimensions, self.factor, self.precision, prefixed
        )

    @property
    def repr(self) -> str:
        """
        Returns a repr that can be used to create another Physical instance.
        """
        repr_str = (
            "Physical(value={}, dimensions={}, factor={}, precision={}, _prefixed={})"
        )
        return repr_str.format(
            self.value, self.dimensions, self.factor, self.precision, self.prefixed
        )  # check

    def round(self, n: int):
        """
        Returns a new Physical with a new precision, 'n'. Precision controls
        the number of decimal places displayed in repr and str.
        """
        return Physical(self.value, self.dimensions, self.factor, n, self.prefixed)

    def split(self, base_value: bool = True) -> tuple:
        """
        Returns a tuple separating the value of `self` with the units of `self`.
        If base_value is True, then the value will be the value in base units. If False, then
        the apparent value of `self` will be used.

        This method is to allow flexibility in working with Physical instances when working
        with numerically optimized libraries such as numpy which cannot accept non-numerical
        objects in some of their operations (such as in matrix inversion).
        """
        if base_value:
            return (
                self.value * self.factor,
                Physical(1 / self.factor, self.dimensions, self.factor, self.precision),
            )
        return (float(self), Physical(1, self.dimensions, self.factor, self.precision))

    def sqrt(self, n: float = 2.0):
        """
        Returns a Physical instance that represents the square root of `self`.
        `n` can be set to an alternate number to compute an alternate root (e.g. 3.0 for cube root)
        """
        return self ** (1 / n)

    def to(self, unit_name=""):
        """
        Returns None and alters the instance into one of the elligible
        alternative units for its dimension, if it exists in the alternative_units dict;
        """
        dims = self.dimensions
        env_dims = environment.units_by_dimension
        derived = env_dims()["derived"]
        defined = env_dims()["defined"]
        power, dims_orig = phf._powers_of_derived(dims, env_dims)
        if not unit_name:
            print("Available units: ")
            for key in derived.get(dims_orig, {}):
                print(key)
            for key in defined.get(dims_orig, {}):
                print(key)

        if unit_name:
            defined_match = defined.get(dims_orig, {}).get(unit_name, {})
            derived_match = derived.get(dims_orig, {}).get(unit_name, {})
            unit_match = defined_match or derived_match
            if not unit_match: warnings.warn(f"No unit defined for '{unit_name}' on {self}.")
            new_factor = unit_match.get("Factor", 1) ** power
            return Physical(self.value, self.dimensions, new_factor, self.precision)

    def si(self):
        """
        Return a new Physical instance with self.factor set to 1, thereby returning
        the instance to SI units display.
        """
        return Physical(self.value, self.dimensions, 1, self.precision)

    ### repr Methods (the "workhorse" of Physical) ###

    def __repr__(self):
        return self._repr_template_()

    def _repr_html_(self):
        return self._repr_template_(template="html")

    def _repr_markdown_(self):
        return self._repr_template_(template="html")

    def _repr_latex_(self):
        return self._repr_template_(template="latex")

    def _repr_template_(self, template: str = "", format_spec=None) -> str:
        """
        Returns a string that appropriately represents the Physical
        instance. The parameter,'template', allows two optional values:
        'html' and 'latex'. which will only be utilized if the Physical
        exists in the Jupyter/iPython environment.
        """
        if format_spec is None:
            format_spec = f".{self.precision}f"
        dims = self.dimensions
        factor = self.factor
        val = self.value
        prefix = ""
        prefixed = self.prefixed
        eps = self._eps
        kg_bool = False

        # Access external environment
        env_fact = environment.units_by_factor or dict()
        env_dims = environment.units_by_dimension or dict()



        # Do the expensive vector math method (call once, only)
        power, dims_orig = phf._powers_of_derived(dims, env_dims)

        # Determine if there is a symbol for these dimensions in the environment
        # and if the quantity is elligible to be prefixed
        symbol, prefix_bool = phf._evaluate_dims_and_factor(
            dims_orig, factor, power, env_fact, env_dims
        )
        # Get the appropriate prefix
        if prefix_bool and prefixed == "unity":
            prefix = ""
        elif prefix_bool and prefixed:
            prefix = prefixed
        elif prefix_bool and dims_orig == Dimensions(1, 0, 0, 0, 0, 0, 0):
            kg_bool = True
            prefix = phf._auto_prefix(val, power, kg=kg_bool)
        elif prefix_bool:
            prefix = phf._auto_prefix(val, power, kg=kg_bool)

        # Format the exponent (may not be used, though)
        exponent = phf._format_exponent(power, repr_format=template, eps=eps)

        # Format the units
        if not symbol and phf._dims_basis_multiple(dims):
            components = phf._get_unit_components_from_dims(dims)
            units_symbol = phf._get_unit_string(components, repr_format=template)
            units = units_symbol
            units = phf._format_symbol(prefix, units_symbol, repr_format=template)
            exponent = ""
        elif not symbol:
            components = phf._get_unit_components_from_dims(dims)
            units_symbol = phf._get_unit_string(components, repr_format=template)
            units = units_symbol
            exponent = ""
        else:
            units = phf._format_symbol(prefix, symbol, repr_format=template)

        # Determine the appropriate display value
        value = val * factor

        if prefix_bool:
            # If the quantity has a "pre-fixed" prefix, it will override
            # the value generated in _auto_prefix_value
            value = phf._auto_prefix_value(val, power, prefix, kg_bool)

        pre_super = ""
        post_super = ""
        space = " "
        if template == "latex":
            space = r"\ "
            pre_super = "^{"
            post_super = "}"
        elif template == "html":
            space = " "
            pre_super = "<sup>"
            post_super = "</sup>"

        if not exponent:
            pre_super = ""
            post_super = ""

        # if not prefix_bool:
        #    return f"{value:.{precision}f}{space}{units}{pre_super}{exponent}{post_super}" 
        return f"{value:{format_spec}}{space}{units}{pre_super}{exponent}{post_super}"

    ### "Magic" Methods ###

    def __float__(self):
        value = self.value
        factor = self.factor
        if factor != 1:
            return value * factor
        kg_bool = False
        dims = self.dimensions
        env_dims = environment.units_by_dimension or dict()
        power, _ = phf._powers_of_derived(dims, env_dims)
        dim_components = phf._get_unit_components_from_dims(dims)
        if len(dim_components) == 1 and dim_components[0][0] == "kg":
            kg_bool = True
        if self.prefixed:
            prefix = self.prefixed
        else:
            prefix = phf._auto_prefix(value, power, kg_bool)
        float_value = phf._auto_prefix_value(value, power, prefix, kg_bool)
        return float_value

    def __int__(self):
        return int(float(self))

    def __neg__(self):
        return self * -1

    def __abs__(self):
        if self.value < 0:
            return self * -1
        return self

    def __bool__(self):
        return True

    def __format__(self, format_spec = ''):
        return self._repr_template_(template="", format_spec=format_spec)

    def __hash__(self):
        return hash(
            (self.value, self.dimensions, self.factor, self.precision, self.prefixed)
        )

    def __round__(self, n=0):
        return self.round(n)

    def __contains__(self, other):
        return False

    def __eq__(self, other):
        if isinstance(other, NUMBER):
            return round(self.value, phf._total_precision) == other
        elif type(other) == str:
            return False
        elif isinstance(other, Physical) and self.dimensions == other.dimensions:
            return round(self.value, phf._total_precision) == round(
                other.value, phf._total_precision
            )
        else:
            raise ValueError(
                "Can only compare between Physical instances of equal dimension."
            )

    def __gt__(self, other):
        if isinstance(other, NUMBER):
            return round(self.value, phf._total_precision) > other
        elif isinstance(other, Physical) and self.dimensions == other.dimensions:
            return round(self.value, phf._total_precision) > round(
                other.value, phf._total_precision
            )
        else:
            raise ValueError(
                "Can only compare between Physical instances of equal dimension."
            )

    def __ge__(self, other):
        if isinstance(other, NUMBER):
            return round(self.value, phf._total_precision) >= other
        elif isinstance(other, Physical) and self.dimensions == other.dimensions:
            return round(self.value, phf._total_precision) >= round(
                other.value, phf._total_precision
            )
        else:
            raise ValueError(
                "Can only compare between Physical instances of equal dimension."
            )

    def __lt__(self, other):
        if isinstance(other, NUMBER):
            return round(self.value, phf._total_precision) < other
        elif isinstance(other, Physical) and self.dimensions == other.dimensions:
            return round(self.value, phf._total_precision) < round(
                other.value, phf._total_precision
            )
        else:
            raise ValueError(
                "Can only compare between Physical instances of equal dimension."
            )

    def __le__(self, other):
        if isinstance(other, NUMBER):
            return round(self.value, phf._total_precision) <= other
        elif isinstance(other, Physical) and self.dimensions == other.dimensions:
            return round(self.value, phf._total_precision) <= round(
                other.value, phf._total_precision
            )
        else:
            raise ValueError(
                "Can only compare between Physical instances of equal dimension."
            )

    def __add__(self, other):
        if isinstance(other, Physical):
            if self.dimensions == other.dimensions:
                try:
                    return Physical(
                        self.value + other.value,
                        self.dimensions,
                        self.factor,
                        self.precision,
                        self.prefixed,
                    )
                except:
                    raise ValueError(
                        f"Cannot add between {self} and {other}: "
                        + ".value attributes are incompatible."
                    )
            else:
                raise ValueError(
                    f"Cannot add between {self} and {other}: "
                    + ".dimensions attributes are incompatible (not equal)"
                )
        else:
            try:
                other = other / self.factor
                return Physical(
                    self.value + other,
                    self.dimensions,
                    self.factor,
                    self.precision,
                    self.prefixed,
                )
            except:
                raise ValueError(
                    f"Cannot add between {self} and {other}: "
                    + ".value attributes are incompatible."
                )

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        raise ValueError(
            "Cannot incrementally add Physical instances because they are immutable."
            + " Use 'a = a + b', to make the operation explicit."
        )

    def __sub__(self, other):
        if isinstance(other, Physical):
            if self.dimensions == other.dimensions:
                try:
                    return Physical(
                        self.value - other.value,
                        self.dimensions,
                        self.factor,
                        self.precision,
                        self.prefixed,
                    )
                except:
                    raise ValueError(f"Cannot subtract between {self} and {other}")
            else:
                raise ValueError(
                    f"Cannot subtract between {self} and {other}:"
                    + ".dimensions attributes are incompatible (not equal)"
                )
        else:
            try:
                other = other / self.factor
                return Physical(
                    self.value - other,
                    self.dimensions,
                    self.factor,
                    self.precision,
                    self.prefixed,
                )
            except:
                raise ValueError(
                    f"Cannot subtract between {self} and {other}: "
                    + ".value attributes are incompatible."
                )

    def __rsub__(self, other):
        if isinstance(other, Physical):
            return self.__sub__(other)
        else:
            try:
                other = other / self.factor
                return Physical(
                    other - self.value,
                    self.dimensions,
                    self.factor,
                    self.precision,
                    self.prefixed,
                )
            except:
                raise ValueError(
                    f"Cannot subtract between {self} and {other}: "
                    + ".value attributes are incompatible."
                )

    def __isub__(self, other):
        raise ValueError(
            "Cannot incrementally subtract Physical instances because they are immutable."
            + " Use 'a = a - b', to make the operation explicit."
        )

    def __mul__(self, other):
        if phf.is_nan(other):
            return other
        elif isinstance(other, NUMBER):
            return Physical(
                self.value * other,
                self.dimensions,
                self.factor,
                self.precision,
                self.prefixed,
            )

        elif isinstance(other, Physical):
            new_dims = vec.add(self.dimensions, other.dimensions)
            new_power, new_dims_orig = phf._powers_of_derived(
                new_dims, environment.units_by_dimension
            )
            new_factor = self.factor * other.factor
            test_factor = phf._get_units_by_factor(
                new_factor, new_dims_orig, environment.units_by_factor, new_power
            )
            if not test_factor:
                new_factor = 1
            try:
                new_value = self.value * other.value
            except:
                raise ValueError(
                    f"Cannot multiply between {self} and {other}: "
                    + ".value attributes are incompatible."
                )
            if new_dims == Dimensions(0, 0, 0, 0, 0, 0, 0):
                return new_value
            else:
                return Physical(new_value, new_dims, new_factor, self.precision)
        else:
            try:
                return Physical(
                    self.value * other, self.dimensions, self.factor, self.precision
                )
            except:
                raise ValueError(
                    f"Cannot multiply between {self} and {other}: "
                    + ".value attributes are incompatible."
                )

    def __imul__(self, other):
        raise ValueError(
            "Cannot incrementally multiply Physical instances because they are immutable."
            + " Use 'a = a * b' to make the operation explicit."
        )

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if phf.is_nan(other):
            return other
        elif isinstance(other, NUMBER):
            return Physical(
                self.value / other,
                self.dimensions,
                self.factor,
                self.precision,
                self.prefixed,
            )
        elif isinstance(other, Physical):
            new_dims = vec.subtract(self.dimensions, other.dimensions)
            new_power, new_dims_orig = phf._powers_of_derived(
                new_dims, environment.units_by_dimension
            )
            new_factor = self.factor / other.factor
            if not phf._get_units_by_factor(
                new_factor, new_dims_orig, environment.units_by_factor, new_power
            ):
                new_factor = 1
            try:
                new_value = self.value / other.value
            except:
                raise ValueError(
                    f"Cannot divide between {self} and {other}: "
                    + ".value attributes are incompatible."
                )
            if new_dims == Dimensions(0, 0, 0, 0, 0, 0, 0):
                return new_value
            else:
                return Physical(new_value, new_dims, new_factor, self.precision)
        else:
            try:
                return Physical(
                    self.value / other, self.dimensions, self.factor, self.precision
                )
            except:
                raise ValueError(
                    f"Cannot divide between {self} and {other}: "
                    + ".value attributes are incompatible."
                )

    def __rtruediv__(self, other):
        if phf.is_nan(other):
            return other
        if isinstance(other, NUMBER):
            new_value = other / self.value
            new_dimensions = vec.multiply(self.dimensions, -1)
            new_factor = self.factor ** -1  # added new_factor
            return Physical(
                new_value,
                new_dimensions,
                new_factor,  # updated from self.factor to new_factor
                self.precision,
            )
        else:
            try:
                return Physical(
                    other / self.value,
                    vec.multiply(self.dimensions, -1),
                    self.factor ** -1,  # updated to ** -1
                    self.precision,
                )
            except:
                raise ValueError(
                    f"Cannot divide between {other} and {self}: "
                    + ".value attributes are incompatible."
                )

    def __itruediv__(self, other):
        raise ValueError(
            "Cannot incrementally divide Physical instances because they are immutable."
            + " Use 'a = a / b' to make the operation explicit."
        )

    def __pow__(self, other):
        if isinstance(other, NUMBER):
            if self.prefixed:
                return float(self) ** other
            new_value = self.value ** other
            new_dimensions = vec.multiply(self.dimensions, other)
            new_factor = self.factor ** other
            return Physical(new_value, new_dimensions, new_factor, self.precision)
        else:
            raise ValueError(
                "Cannot raise a Physical to the power of \
                                     another Physical -> ({self}**{other})".format(
                    self, other
                )
            )


# The seven SI base units...
_the_si_base_units = {
    "kg": Physical(1, Dimensions(1, 0, 0, 0, 0, 0, 0), 1.0),
    "m": Physical(1, Dimensions(0, 1, 0, 0, 0, 0, 0), 1.0),
    "s": Physical(1, Dimensions(0, 0, 1, 0, 0, 0, 0), 1.0),
    "A": Physical(1, Dimensions(0, 0, 0, 1, 0, 0, 0), 1.0),
    "cd": Physical(1, Dimensions(0, 0, 0, 0, 1, 0, 0), 1.0),
    "K": Physical(1, Dimensions(0, 0, 0, 0, 0, 1, 0), 1.0),
    "mol": Physical(1, Dimensions(0, 0, 0, 0, 0, 0, 1), 1.0),
}

environment = Environment(Physical, builtins, _the_si_base_units)
environment.push_vars(_the_si_base_units, sys.modules[__name__])




#   Copyright 2020 Connor Ferster

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math
import pytest
import forallpeople as si
import forallpeople.physical_helper_functions as phf

si.environment("test_definitions", top_level = True)


### TODO: add Ohms to definitions for testing
# add energy modelling units to definitions for testing

### Testing parameters ###
env_dims = si.environment.units_by_dimension
env_fact = si.environment.units_by_factor
units = {
    "A": 0.05 * kg,
    "B": 3.2e-3 * m,
    "C": 1000 * ft,
    "D": 1e6 * N,
    "E": 0.2 * kip,
    "F": 5 * N * 1e3 * kip,
    "Matt": 0.22 * kip,  ## 20200417Matt added for dummy merge test
}
parameters = [
    (value, phf._powers_of_derived(value.dimensions, env_dims))
    for value in units.values()
]

### Tests of the Physical class ###

## Testing "._repr_" methods in order of appearance in ._repr_template_() ##
def test__evaluate_dims_and_factor():
    func = phf._evaluate_dims_and_factor
    assert func(
        si.Dimensions(1, 1, -2, 0, 0, 0, 0),
        1 / 0.45359237 / 9.80665,
        1,
        env_fact,
        env_dims,
    ) == ("lb", False)
    assert func(si.Dimensions(1, 1, -2, 0, 0, 0, 0), 1, 2, env_fact, env_dims) == (
        "N",
        True,
    )
    assert func(si.Dimensions(1, 1, -2, 0, 0, 0, 0), 1, 1, env_fact, env_dims) == (
        "N",
        True,
    )
    assert func(si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1 / 0.3048, 1, env_fact, env_dims)
    assert func(si.Dimensions(1, 0, 0, 0, 0, 0, 0), 1, 3, env_fact, env_dims) == (
        "",
        True,
    )
    assert func(si.Dimensions(1, 1, 1, 0, 0, 0, 0), 1, 1, env_fact, env_dims) == (
        "",
        False,
    )


def test__get_units_by_factor():
    ftlb = lb * ft
    ft2 = ft ** 2
    func = phf._get_units_by_factor
    assert func(ft.factor, ft.dimensions, env_fact, 1) == {
        "ft": {
            "Dimension": si.Dimensions(kg=0, m=1, s=0, A=0, cd=0, K=0, mol=0),
            "Symbol": "ft",
            "Factor": 3.280839895013123,
        }
    }
    assert func(ft2.factor, ft.dimensions, env_fact, 2) == {
        "ft": {
            "Dimension": si.Dimensions(kg=0, m=1, s=0, A=0, cd=0, K=0, mol=0),
            "Symbol": "ft",
            "Factor": 3.280839895013123,
        }
    }
    assert func(ftlb.factor, ftlb.dimensions, env_fact, 1) == {
        "lbft": {
            "Dimension": si.Dimensions(kg=1, m=2, s=-2, A=0, cd=0, K=0, mol=0),
            "Symbol": "lb·ft",
            "Factor": 0.7375621492772653,
        }
    }
    assert func((ftlb * ft).factor, (ftlb * ft).dimensions, env_fact, 1) == dict()


def test__get_unit_components_from_dims():
    func = phf._get_unit_components_from_dims
    assert func(si.Dimensions(1, 1, -2, 0, 0, 0, 0)) == [("kg", 1), ("m", 1), ("s", -2)]
    assert func(si.Dimensions(1, 2, 3, 4, 5, 6, 7)) == [
        ("kg", 1),
        ("m", 2),
        ("s", 3),
        ("A", 4),
        ("cd", 5),
        ("K", 6),
        ("mol", 7),
    ]
    assert func(si.Dimensions(0, 0, 0, 0, 0, 0, 0)) == []


# TODO: Complete these tests

# def test__format_symbol():
#     assert False


# def test__format_exponent():
#     assert False

def test_nan_pass_through():
    assert math.isnan(5*kg*float('nan'))
    assert math.isnan(float('nan')*ft)


def test__get_unit_string():
    func = phf._get_unit_string
    assert func([("kg", 1), ("m", 1), ("s", -2)], "") == "kg·m·s⁻²"
    assert (
        func([("kg", 1), ("m", 1), ("s", -2)], "html")
        == "kg&#8901;m&#8901;s<sup>-2</sup>"
    )
    assert (
        func([("kg", 1), ("m", 1), ("s", -2)], "latex")
        == r"\text{kg} \cdot \text{m} \cdot \text{s}^{-2}"
    )
    assert func([("m", 2)], "html") == "m<sup>2</sup>"


def test__get_superscript_string():
    func = phf._get_superscript_string
    assert func("-2") == "⁻²"
    assert func("1.39") == "¹'³⁹"
    assert func("-0.10") == "⁻⁰'¹⁰"


def test_latex():
    assert MPa.latex == "1.000\\ \\text{MPa}"
    assert (2.5 * kg * m ** 2.5).latex == "2.500\\ \\text{kg} \\cdot \\text{m}^{2.5}"


def test_repr():
    assert (
        MPa.repr
        == "Physical(value=1000000.0, dimensions="
        + "Dimensions(kg=1, m=-1, s=-2, A=0, cd=0, K=0, mol=0), "
        + "factor=1, precision=3, _prefixed=)"
    )
    assert (
        ft.repr
        == "Physical(value=0.3048, dimensions="
        + "Dimensions(kg=0, m=1, s=0, A=0, cd=0, K=0, mol=0), "
        + "factor=3.280839895013123, precision=3, _prefixed=)"
    )


def test_html():
    assert (25 * m ** 2).html == "25.000 m<sup>2</sup>"
    assert (units["A"]).html == "50.000 g"


def test_prefixed():
    assert (25*m**2).prefix("c").html == "250000.000 cm<sup>2</sup>"


def test_round():
    assert repr((25.2398783 * N).round(4)) == "25.2399 N"


def test_split():
    assert units["C"].split()[0] == pytest.approx(1000)
    assert units["C"].split()[1] == ft


def test_sqrt():
    assert (25 * m / s).sqrt() == 5 * (m ** 0.5 / (s ** 0.5))


def test_in_units():
    # assert ft.to('m').factor == 1
    assert m.to("ft").factor == 1 / 0.3048
    assert kip.to("lb").factor == (1000 * lb).factor
    assert ((10 * lb) ** 2).to("kip").factor == (0.1 * kip * kip).factor


def test__check_dims_parallel():
    assert phf._check_dims_parallel((3, 2, 1, 0), (6, 4, 2, 0))
    assert phf._check_dims_parallel(
        si.Dimensions(1, 2, -2, 0, 0, 0, 0), si.Dimensions(-1, -2, 2, 0, 0, 0, 0)
    )


def test__get_derived_unit():
    func = phf._get_derived_unit
    assert func(si.Dimensions(1, 1, -2, 0, 0, 0, 0), env_dims) == {
        "N": {
            "Dimension": si.Dimensions(kg=1, m=1, s=-2, A=0, cd=0, K=0, mol=0),
            "Factor": 1,
        }
    }
    assert func(si.Dimensions(1, 0, 0, 0, 0, 0, 0), env_dims) == {}
    assert func(si.Dimensions(1, 0, -2, 0, 0, 0, 0), env_dims) == {
        "N_m": {
            "Dimension": si.Dimensions(kg=1, m=0, s=-2, A=0, cd=0, K=0, mol=0),
            "Factor": 1,
            "Symbol": "N/m",
        }
    }
    assert func(si.Dimensions(1, 1, 1, 1, 1, 1, 1), env_dims) == {}


def test__dims_quotient():
    func = phf._dims_quotient
    assert func(si.Dimensions(1, 1, -2, 0, 0, 0, 0), env_dims) == si.Dimensions(
        1, 1, 1, 0, 0, 0, 0
    )
    assert func(si.Dimensions(3, 3, -6, 0, 0, 0, 0), env_dims) == si.Dimensions(
        3, 3, 3, 0, 0, 0, 0
    )
    assert func(si.Dimensions(1, 2, -2, 0, 0, 0, 0), env_dims) == si.Dimensions(
        1, 1, 1, 0, 0, 0, 0
    )


def test__dims_basis_multiple():
    func = phf._dims_basis_multiple
    assert func(si.Dimensions(0, 1, 0, 0, 0, 0, 0)) == si.Dimensions(
        0, 1, 0, 0, 0, 0, 0
    )
    assert func(si.Dimensions(0, 4, 0, 0, 0, 0, 0)) == si.Dimensions(
        0, 4, 0, 0, 0, 0, 0
    )
    assert func(si.Dimensions(0, 0, 2.5, 0, 0, 0, 0)) == si.Dimensions(
        0, 0, 2.5, 0, 0, 0, 0
    )
    assert func(si.Dimensions(0, 1, 2.5, 0, 0, 0, 0)) == None
    assert func(si.Dimensions(1, 1, -2, 0, 0, 0, 0)) == None


def test__powers_of_derived():
    func = phf._powers_of_derived
    dims = si.Dimensions
    assert func(dims(0, 1, 0, 0, 0, 0, 0), env_dims) == (1, dims(0, 1, 0, 0, 0, 0, 0))
    assert func(dims(1, 1, -2, 0, 0, 0, 0), env_dims) == (1, dims(1, 1, -2, 0, 0, 0, 0))
    assert func(dims(3, 3, -6, 0, 0, 0, 0), env_dims) == (3, dims(1, 1, -2, 0, 0, 0, 0))
    assert func(dims(1, 2, -2, 0, 0, 0, 0), env_dims) == (1, dims(1, 2, -2, 0, 0, 0, 0))
    assert func(dims(0, 4, 0, 0, 0, 0, 0), env_dims) == (4, dims(0, 1, 0, 0, 0, 0, 0))
    assert func(dims(0, 0, 2.5, 0, 0, 0, 0), env_dims) == (
        2.5,
        dims(0, 0, 1, 0, 0, 0, 0),
    )
    assert func(dims(3.6, 3.6, -7.2, 0, 0, 0, 0), env_dims) == (
        3.6,
        dims(1, 1, -2, 0, 0, 0, 0),
    )


def test__auto_prefix():
    func = phf._auto_prefix
    assert func(1500, 1) == "k"
    assert func(1500, 2) == ""
    assert func(1.5e6, 2) == "k"
    assert func(1.5e6, 1) == "M"
    assert func(1.5e-3, 1) == "m"
    assert func(1.5e-6, 2) == "m"


def test__auto_prefix_kg():
    func = phf._auto_prefix_kg
    assert func(1500, 1) == "M"
    assert func(1500, 2) == "k"
    assert func(1.5e6, 2) == "M"
    assert func(1.5e6, 1) == "G"
    assert func(1.5e-3, 1) == ""
    assert func(1.5e-6, 2) == ""
    assert func(1.5e-6, 1) == "m"


def test__auto_prefix_value():
    func = phf._auto_prefix_value
    assert func(1500, 1, "k") == 1.5
    assert func(1500, 2, "") == 1500
    assert func(52500, 1, "k") == 52.5
    assert func(1.5e6, 2, "k") == 1.5
    assert func(1.5e6, 1, "M") == 1.5
    assert func(1.5e-6, 2, "m") == 1.5
    assert func(1.5e-5, 1, "μ") == pytest.approx(15)
    assert func(1.5e-5, 2, "m") == pytest.approx(15)


def test___eq__():
    assert m == m
    assert m == 1
    assert N == 1
    assert Pa == 1
    assert kg == 1
    with pytest.raises(ValueError):
        kg == m
        N == Pa


def test___gt__():
    assert m > ft
    assert MPa > Pa
    assert (5 * m > 5 * m) == False
    assert 5 * kip > 30 * N
    with pytest.raises(ValueError):
        5 * kg > 3 * m


def test___ge__():
    assert m >= ft
    assert MPa >= Pa
    assert 5 * m >= 5 * m
    assert 5 * kN >= 5000 * N
    with pytest.raises(ValueError):
        5 * kg >= 5 * m


def test___lt__():
    assert ft < m
    assert Pa < MPa
    assert (5 * m < 5 * m) == False
    assert 5 * N < 5001 * kN
    with pytest.raises(ValueError):
        5 * kg < 5 * m


def test___le__():
    assert ft <= m
    assert Pa <= MPa
    assert 5 * m <= 5 * m
    assert 5000 * N <= 5 * kN
    with pytest.raises(ValueError):
        5 * kg <= 5 * m


def test___add__():
    assert kg + kg == si.Physical(2, si.Dimensions(1, 0, 0, 0, 0, 0, 0), 1)
    assert m + ft == si.Physical(1.3048, si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1)
    assert ft + m == si.Physical(1.3048, si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1 / 0.3048)
    assert N + lb == si.Physical(
        5.4482216152605005, si.Dimensions(1, 1, -2, 0, 0, 0, 0), 1
    )
    assert lb + N == si.Physical(
        5.4482216152605005, si.Dimensions(1, 1, -2, 0, 0, 0, 0), 0.22480894309971047
    )
    assert ft + 3 == si.Physical(1.2192, si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1 / 0.3048)
    with pytest.raises(ValueError):
        kg + m
        N + psf


# def test___iadd__():
#     with pytest.raises(ValueError):
#         m += 3


def test___sub__():
    assert kg - kg == 0.0
    assert m - ft == si.Physical(0.6952, si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1)
    assert ft - m == si.Physical(
        -0.6952, si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1 / 0.3048
    )
    assert N - lb == si.Physical(
        -3.4482216152605005, si.Dimensions(1, 1, -2, 0, 0, 0, 0), 1
    )
    assert lb - N == si.Physical(
        3.4482216152605005, si.Dimensions(1, 1, -2, 0, 0, 0, 0), 0.22480894309971047
    )
    assert (ft - 3).value == pytest.approx(
        si.Physical(-0.6096, si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1 / 0.3048).value
    )
    with pytest.raises(ValueError):
        kg - m
        N - psf


def test___rsub__():
    assert 2 - ft == si.Physical(0.3048, si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1 / 0.3048)
    assert 10 - N == si.Physical(9, si.Dimensions(1, 1, -2, 0, 0, 0, 0), 1)


# def test___isub__():
#     with pytest.raises(ValueError):
#         m -= 3


def test___mul__():
    assert m * m == si.Physical(1, si.Dimensions(0, 2, 0, 0, 0, 0, 0), 1)
    assert m * kg == si.Physical(1, si.Dimensions(1, 1, 0, 0, 0, 0, 0), 1)
    assert ft * m == si.Physical(0.3048, si.Dimensions(0, 2, 0, 0, 0, 0, 0), 1 / 0.3048)
    assert N * m == si.Physical(1, si.Dimensions(1, 2, -2, 0, 0, 0, 0), 1)
    assert psf * m * m == si.Physical(
        47.88025898033584, si.Dimensions(1, 1, -2, 0, 0, 0, 0), 0.02088543423315013
    )
    assert 2 * ft == si.Physical(0.6096, si.Dimensions(0, 1, 0, 0, 0, 0, 0), 1 / 0.3048)
    assert (
        si.Physical(10, si.Dimensions(-1, -1, 0, 0, 0, 0, 0), 1)
        * si.Physical(2, si.Dimensions(1, 1, 0, 0, 0, 0, 0), 1)
    ) == 20
    assert (10 * ksf) * (5 * ft) * (2 * ft) == 100 * kip  # TODO: Fix this gotcha
    assert m * ft == si.Physical(0.3048, si.Dimensions(0, 2, 0, 0, 0, 0, 0), 1.0)


# def test___imul__():
#     with pytest.raises(ValueError):
#         m *= 3


def test___truediv__():
    assert m / m == 1
    assert m / s == si.Physical(1, si.Dimensions(0, 1, -1, 0, 0, 0, 0), 1)
    assert kN / m / m == kPa
    assert (5 * kN) / 2 == 2.5 * kN
    assert (kip / (2 * ft * 5 * ft)).value == pytest.approx((0.100 * ksf).value)


def test___rtruediv__():
    assert 2 / m == si.Physical(2, si.Dimensions(0, -1, 0, 0, 0, 0, 0), 1)
    assert 10 / N == si.Physical(10, si.Dimensions(-1, -1, 2, 0, 0, 0, 0), 1)
    assert 1 / kip == si.Physical(
        value=0.00022480894309971045,
        dimensions=si.Dimensions(kg=-1, m=-1, s=2, A=0, cd=0, K=0, mol=0),
        factor=4448.221615260501,
        precision=3,
    )


def test___pow__():
    assert N ** 2 == si.Physical(1, si.Dimensions(2, 2, -4, 0, 0, 0, 0), 1)
    assert ft ** 3 == si.Physical(
        0.3048 ** 3, si.Dimensions(0, 3, 0, 0, 0, 0, 0), (1 / 0.3048) ** 3
    )


def test___abs__():
    assert abs(-1 * m) == m
    assert abs(-10 * ft ** 2) == 10 * ft ** 2
    assert abs(1 * kip) == 1 * kip


def test___float__():
    assert float(6.7 * ft) == pytest.approx(6.7)
    assert float((52.5 * kN)) == pytest.approx(52.5)
    assert float(34 * kg * A * m) == pytest.approx(34)


## Test of Environment Class ##

# def test_load_environment():
#    assert "stub" == False
#
# def test_instantiator():
#    assert "stub" == False


## Tests of supplementary math functions ##


def test_sqrt():
    assert (9 * kPa).sqrt() == 3 * kPa ** 0.5
    assert (9 * MPa).sqrt() == 3 * MPa ** 0.5


## Integration tests
def test_defined_unit_persistence():
    a = units["E"]
    x = units["D"]
    b = 1 / a ** 2

    area = 2.3 * ft * 1.2 * ft
    c = a / area

    d = a.to("lb")
    e = 2.8 * lb
    f = 1 / d ** 2 * e * 10
    g = 1 / f + x / 1e3

    assert repr(b) == "25.000 kip⁻²"
    assert (
        b.repr
        == "Physical(value=1.2634765224402213e-06, dimensions=Dimensions(kg=-2, m=-2, s=4, A=0, cd=0, K=0, mol=0), factor=19786675.538470734, precision=3, _prefixed=)"
    )
    assert repr(c) == "0.072 ksf"
    assert repr(g) == "1653.380 lb"



#   Copyright 2020 Connor Ferster

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import annotations
from collections import ChainMap
import functools
import math
from typing import Any, Union, Optional, List, Callable
from forallpeople.dimensions import Dimensions
import forallpeople.tuplevector as vec

### Helper methods for repr methods ###

_prefixes = { # Do not add custom prefixes between Y and y, e.g. "c": 1e-2
    "Y": 1e24,
    "Z": 1e21,
    "E": 1e18,
    "P": 1e15,
    "T": 1e12,
    "G": 1e09,
    "M": 1e06,
    "k": 1e03,
    "": 1.0,
    "m": 1e-3,
    "μ": 1e-6,
    "n": 1e-09,
    "p": 1e-12,
    "f": 1e-15,
    "a": 1e-18,
    "z": 1e-21,
    "y": 1e-24,
}

_prefix_lookups = { # Do not add custom prefixes between Y and y, e.g. "c": 1e-2
    24:"Y", 
    21:"Z", 
    18:"E", 
    15:"P", 
    12:"T", 
    9:"G", 
    6:"M", 
    3:"k", 
    0: "",
    -3: "m", 
    -6: "μ", 
    -9: "n",
    -12: "p",
    -15: "f",
    -18: "a",
    -21: "z",
    -24: "y",
}

_additional_prefixes = {
    "c": 1e-2,
}

_superscripts = {
    "1": "¹",
    "2": "²",
    "3": "³",
    "4": "⁴",
    "5": "⁵",
    "6": "⁶",
    "7": "⁷",
    "8": "⁸",
    "9": "⁹",
    "0": "⁰",
    "-": "⁻",
    ".": "'",
}
_eps = 1e-7
_total_precision = 6

@functools.lru_cache(maxsize=None)
def _evaluate_dims_and_factor(
    dims_orig: Dimensions,
    factor: Union[int, float],
    power: Union[int, float],
    env_fact: Callable,
    env_dims: Callable,
) -> tuple:
    """Part of the __str__ and __repr__ process.
    Returns a tuple containing the
    appropriate symbol as a string (if applicable; '' if not) and a
    boolean indicating whether or not the dimension and factor combination
    is elligible for a prefix."""
    defined = _get_units_by_factor(
        factor=factor, dims=dims_orig, units_env=env_fact, power=power
    )

    # Derived units not retrieving inverted definitions
    derived = _get_derived_unit(dims=dims_orig, units_env=env_dims)
    single_dim = _dims_basis_multiple(dims_orig)
    if defined:
        units_match = defined
        prefix_bool = False
    elif derived or single_dim:
        units_match = derived
        prefix_bool = True
    else:
        units_match = derived
        prefix_bool = False

    if units_match:
        name = tuple(units_match.keys())[0]
        symbol = units_match.get(name, {}).get("Symbol", "")
        symbol = symbol or name
    else:
        symbol = ""
    return (symbol, prefix_bool)

@functools.lru_cache(maxsize=None)
def _get_units_by_factor(
    factor: float, dims: Dimensions, units_env: Callable, power: Union[int, float]
) -> dict:
    """
    Returns a units_dict from the environment instance if the numerical
    value of 'factor' is a match for a derived unit defined in the
    environment instance and the dimensions stored in the units_dict are
    equal to 'dims'. Returns an empty dict, otherwise.
    """
    new_factor = factor ** (1 / power)
    units_match = units_env().get(round(new_factor, _total_precision), dict())
    try:
        units_name = tuple(units_match.keys())[0]
    except IndexError:
        units_name = ""
    retrieved_dims = units_match.get(units_name, dict()).get("Dimension", dict())
    if dims != retrieved_dims:
        return dict()
    return units_match


def _get_derived_unit(dims: Dimensions, units_env: dict) -> dict:
    """
    Returns a units definition dict that matches 'dimensions'.
    If 'dimensions' is a derived unit raised to a power (e.g. N**2),
    then its original dimensions are checked instead of the altered ones.
    Returns {} if no unit definition matches 'dimensions'.
    """
    derived_units = units_env().get("derived")
    return derived_units.get(dims, dict())

def _get_unit_string(unit_components: list, repr_format: str) -> str:
    """
    Part of the __str__ and __repr__ process. Returns a string representing
    the SI unit components of the Physical instance extracted from the list of
    tuples, 'unit_components', using 'repr_format' as given by the _repr_x_
    function it was called by. If 'repr_format' is not given, then terminal
    output is assumed.
    """
    dot_operator = "·"  # new: · , # old: ⋅
    pre_super = ""
    post_super = ""
    pre_symbol = ""
    post_symbol = ""
    if repr_format == "html":
        dot_operator = "&#8901;"  # &#183;
        pre_super = "<sup>"
        post_super = "</sup>"
    elif repr_format == "latex":
        dot_operator = r" \cdot "
        pre_symbol = "\\text{"
        post_symbol = "}"
        pre_super = "^{"
        post_super = "}"

    str_components = []
    kg_only = ""
    for symbol, exponent in unit_components:
        if exponent:
            kg_only = symbol
        if exponent == 1:
            this_component = f"{pre_symbol}{symbol}{post_symbol}"
        else:
            if not repr_format:
                exponent = _get_superscript_string(str(exponent))
            this_component = (
                f"{pre_symbol}{symbol}{post_symbol}"
                f"{pre_super}{exponent}{post_super}"
            )
        str_components.append(this_component)
    if kg_only == "kg":  # Hack for lone special case of a kg only Physical
        return dot_operator.join(str_components).replace("kg", "g")
    return dot_operator.join(str_components)


@functools.lru_cache(maxsize=None)
def _get_unit_components_from_dims(dims: Dimensions):
    """
    Returns a list of tuples to represent the current units based
    on the current dimensions. Dimension ignored if 0.
    e.g. [('kg', 1), ('m', -1), ('s', -2)]
    """
    unit_components = []
    unit_symbols = dims._fields
    for idx, dim in enumerate(dims):
        if dim:  # int
            unit_tuple = (unit_symbols[idx], dim)
            unit_components.append(unit_tuple)
    return unit_components


@functools.lru_cache(maxsize=None)
def _format_symbol(prefix: str, symbol: str, repr_format: str = "") -> str:
    """
    Returns 'symbol' formatted appropriately for the 'repr_format' output.
    """
    # if r"\text" or "^" in symbol: # in case pre-formatted latex from unit_string
    #    return symbol
    symbol_string_open = ""
    symbol_string_close = ""
    dot_operator = "·"
    ohm = "Ω"
    if repr_format == "html":
        dot_operator = "&#8901;"
        ohm = "&#0937;"
    elif repr_format == "latex":
        dot_operator = " \\cdot "
        ohm = "$\\Omega$"
        symbol_string_open = "\\text{"
        symbol_string_close = "}"

    symbol = (
        symbol.replace("·", symbol_string_close + dot_operator + symbol_string_open)
        .replace("*", symbol_string_close + dot_operator + symbol_string_open)
        .replace("Ω", ohm)
    )
    formatted_symbol = f"{symbol_string_open}{prefix}{symbol}{symbol_string_close}"
    if symbol.startswith("\\text{"):  # special case for 'single dimension' Physicals...
        formatted_symbol = f"{symbol[0:6]}{prefix}{symbol[6:]}"
    return formatted_symbol


def _format_exponent(
    power: Union[int, float], repr_format: str = "", eps: float = 1e-7
) -> str:
    """
    Returns the number in 'power' as a formatted exponent for text display.
    """
    if power == 1:
        return ""

    if abs((abs(power) - round(abs(power)))) <= eps:
        power = int(round(power))
    exponent = str(power)
    if not repr_format:
        exponent = _get_superscript_string(exponent)
    return exponent


def _get_superscript_string(exponent: str) -> str:
    """Part of the __str__ and __repr__ process. Returns the unicode
    "superscript" equivalent string for a given float."""
    exponent_components = list(exponent)
    exponent_string = ""
    for component in exponent_components:
        exponent_string += _superscripts[component]
    return exponent_string


### Mathematical helper functions ###

@functools.lru_cache(maxsize=None)
def _powers_of_derived(dims: Dimensions, units_env: Callable) -> Union[int, float]:
    """
    Returns an integer value that represents the exponent of a unit if the
    dimensions
    array is a multiple of one of the defined derived units in dimension_keys.
    Returns None,
    otherwise.
    e.g. a force would have dimensions = [1,1,-2,0,0,0,0] so a Physical object
    that had dimensions = [2,2,-4,0,0,0,0] would really be a force to the power of
    2.
    This function returns the 2, stating that `dims` is the second power of a
    derived dimension in `units_env`.
    """
    quotient_1 = _dims_quotient(dims, units_env)
    quotient_2 = _dims_basis_multiple(dims)
    quotient_1_mean = None
    if quotient_1 is not None:
        quotient_1_mean = cache_vec_mean(quotient_1, ignore_empty=True)
        
    if quotient_1 is not None and quotient_1_mean != -1:
        power_of_derived = cache_vec_mean(quotient_1, ignore_empty=True)
        base_dimensions = cache_vec_divide(dims, quotient_1, ignore_zeros=True)
        return ((power_of_derived or 1), base_dimensions)
    elif quotient_1_mean == -1 and quotient_2 is not None: # Situations like Hz and s
        power_of_basis = cache_vec_mean(quotient_2, ignore_empty=True)
        base_dimensions = cache_vec_divide(dims, quotient_2, ignore_zeros=True)
        return ((power_of_basis or 1), base_dimensions)
    elif quotient_1_mean == -1: # Now we can proceed with an inverse  unit
        power_of_derived = cache_vec_mean(quotient_1, ignore_empty=True)
        base_dimensions = cache_vec_divide(dims, quotient_1, ignore_zeros=True)
        return ((power_of_derived or 1), base_dimensions)
    elif quotient_2 is not None:
        power_of_basis = cache_vec_mean(quotient_2, ignore_empty=True)
        base_dimensions = cache_vec_divide(dims, quotient_2, ignore_zeros=True)
        return ((power_of_basis or 1), base_dimensions)
    else:
        return (1, dims)

@functools.lru_cache(maxsize=None)
def _dims_quotient(dimensions: Dimensions, units_env: Callable) -> Optional[Dimensions]:
    """
    Returns a Dimensions object representing the element-wise quotient between
    'dimensions' and a defined unit if 'dimensions' is a scalar multiple
    of a defined unit in the global environment variable.
    Returns None otherwise.
    """
    derived = units_env()["derived"]
    defined = units_env()["defined"]
    all_units = ChainMap(defined, derived)
    potential_inv = None # A flag to catch a -1 value (an inversion)
    quotient = None
    quotient_result = None
    for dimension_key in all_units.keys():
        if _check_dims_parallel(dimension_key, dimensions):
            quotient = cache_vec_divide(dimensions, dimension_key, ignore_zeros=True)
            mean = cache_vec_mean(quotient, ignore_empty=True)
            if mean == -1: 
                potential_inv = quotient
            elif -1 < mean < 1:
                return None # Ignore parallel dimensions if they are fractional dimensions
            else:
                quotient_result = quotient
    return quotient_result or potential_inv # Inversion ok, if only option


@functools.lru_cache(maxsize=None)
def cache_vec_divide(tuple_a, tuple_b, ignore_zeros):
    """
    Wraps vec.divide with an lru_cache
    """
    return vec.divide(tuple_a, tuple_b, ignore_zeros)


@functools.lru_cache(maxsize=None)
def cache_vec_mean(tuple_a, ignore_empty):
    """
    Wraps vec.mean with an lru_cache
    """
    return vec.mean(tuple_a, ignore_empty)  


@functools.lru_cache(maxsize=None)
def _check_dims_parallel(d1: Dimensions, d2: Dimensions) -> bool:
    """
    Returns True if d1 and d2 are parallel vectors. False otherwise.
    """
    return vec.multiply(d1, vec.dot(d2, d2)) == vec.multiply(d2, vec.dot(d1, d2))


def _dims_basis_multiple(dims: Dimensions) -> Optional[Dimensions]:
    """
    Returns `dims` if `dims` is a scalar multiple of one of the basis vectors.
    Returns None, otherwise.
    This is used as a check to see if `dims` contains only a single dimension,
    even if that single dimension is to a higher power.
    e.g.
    if `dims` equals Dimensions(2, 0, 0, 0, 0, 0, 0) then `dims` will be
    returned.
    if `dims` equals Dimensions(0, 1, 1, 0, 0, 0, 0) then None will be returned.
    if `dims` equals Dimensions(0, 14, 0, 0, 0, 0, 0) then `dims` will be returned.
    """
    count = 0
    for dim in dims:
        if dim:
            count += 1
        if count > 1:
            return None
    return dims


def _auto_prefix(value: float, power: Union[int, float], kg: bool = False) -> str:
    """
    Returns a string "prefix" of an appropriate value if self.value should be prefixed
    i.e. it is a big enough number (e.g. 5342 >= 1000; returns "k" for "kilo")
    """
    if value == 0: return ""
    kg_factor = 0
    if kg:
        kg_factor = 3
    prefixes = _prefixes
    abs_val = abs(value)
    value_power_of_ten = math.log10(abs_val)
    value_power_of_1000 = value_power_of_ten // (3 * power)
    prefix_power_of_1000 = value_power_of_1000 * 3 + kg_factor
    try: 
        return _prefix_lookups[prefix_power_of_1000]
    except KeyError:
        return None


def _auto_prefix_kg(value: float, power: Union[int, float]) -> str:
    """
    Just like _auto_prefix but handles the one special case for "kg" because it already
    has a prefix of "k" as an SI base unit. The difference is the comparison of
    'power_of_ten'/1000 vs 'power_of_ten'.
    """
    prefixes = _prefixes
    if abs(value) >= 1:
        for prefix, power_of_ten in prefixes.items():
            if abs(value) >= (power_of_ten / 1000) ** abs(power):
                return prefix
    else:
        reverse_prefixes = sorted(prefixes.items(), key=lambda prefix: prefix[0])
        # Get the smallest prefix to start...
        previous_prefix = reverse_prefixes[0][0]
        for prefix, power_of_ten in reversed(list(prefixes.items())):
            if abs(value) < (power_of_ten / 1000) ** abs(power):
                return previous_prefix
            else:
                previous_prefix = prefix


def _auto_prefix_value(
    value: float, power: Union[int, float], prefix: str, kg_bool = False,
) -> float:
    """
    Converts the value to a prefixed value if the instance has a symbol defined in
    the environment (i.e. is in the defined units dict)
    """
    kg_factor = 1
    if kg_bool:
        kg_factor = 1000
    if prefix in _additional_prefixes:
        return value / ((_additional_prefixes[prefix] / kg_factor) ** power)
    if 0 < value < 1:
        return value / ((_prefixes[prefix] / kg_factor ) ** abs(power))
    return value / ((_prefixes[prefix] / kg_factor) ** power)




# def _auto_prefix_value(
#     value: float, power: Union[int, float], prefixed: str = "", kg: bool = False,
# ) -> float:
#     """
#     Converts the value to a prefixed value if the instance has a symbol defined in
#     the environment (i.e. is in the defined units dict)
#     """
#     abs_val = abs(value)
#     if prefixed == "unity": return value
#     kg_factor = 1
#     if kg:
#         kg_factor = 1000
#     if prefixed in _additional_prefixes: prefixes = _additional_prefixes
#     else: prefixes = _prefixes
#     if prefixed:
#         return value / ((prefixes[prefixed] / kg_factor) ** power)


#     if abs_val >= 1:
        
#         for prefix, power_of_ten in prefixes.items():
#             if abs_val >= (power_of_ten / kg_factor) ** abs(power):
#                 return value / ((power_of_ten / kg_factor) ** power)
#     else:
#         reverse_prefixes = sorted(prefixes.items(), key=lambda pre_fact: pre_fact[1])
#         # Get the smallest factor to start...
#         previous_power_of_ten = reverse_prefixes[0][1]
#         for prefix, power_of_ten in reversed(list(prefixes.items())):
#             if abs_val < (power_of_ten / kg_factor) ** abs(power):
#                 return value / ((previous_power_of_ten / kg_factor) ** abs(power))
#             else:
#                 previous_power_of_ten = power_of_ten


def swap_scientific_notation_float(value: float, precision: int) -> str:
    """
    Returns a deque representing 'pycode_as_deque' with any python floats that
    will get "cut-off" by the 'precision' arg when they are rounded as being 
    rendered as strings in python's "e format" scientific notation.

    A float is "cut-off" by 'precision' when it's number of significant digits will
    be less than those required by precision. 

    e.g. elem = 0.001353 with precision=3 will round to 0.001, with only one
    significant digit (1 < 3). Therefore this float is "cut off" and will be 
    formatted instead as "1.353e-3"

    elem = 0.1353 with precision=3 will round to 0.135 with three significant digits
    (3 == 3). Therefore this float will not be formatted.
    """
    if test_for_small_float(value, precision):
        new_value = (
                "{:.{precision}e}".format(value, precision=precision)
                .replace("e-0", "e-")
                .replace("e+0", "e+")
            )
        return new_value
    return


def test_for_small_float(value: float, precision: int) -> bool:
    """
    Returns True if 'value' is a float whose rounded str representation
    has fewer significant figures than the number in 'precision'. 
    Return False otherwise.
    """
    if not isinstance(value, (float)):
        return False
    if value == 0:
        return False
    value_as_str = str(round(abs(value), precision))
    if "e" in str(value):
        return True
    if "." in value_as_str:
        left, *_right = value_as_str.split(".")
        if left != "0":
            return False
    if (
        round(value, precision) != round(value, precision + 1)
        or str(abs(round(value, precision))).replace("0", "").replace(".", "")
        == str(abs(round(value, precision + 1))).replace("0", "").replace(".", "")
        == ""
    ):
        return True
    else:
        return False


def swap_scientific_notation_str(value_as_str: str) -> str:
    """
    Returns a deque representing 'line' with any python 
    float elements in the deque
    that are in scientific notation "e" format converted into a Latex 
    scientific notation.
    """
    b = "}"
    if test_for_scientific_notation_str(value_as_str):
        new_value_as_str = value_as_str.replace("e", " \\times 10 ^ {") + b
        return new_value_as_str
    return value_as_str

def test_for_scientific_notation_str(value_as_str: str) -> bool:
    """
    Returns True if 'elem' represents a python float in scientific
    "e notation".
    e.g. 1.23e-3, 0.09e5
    Returns False otherwise
    """
    test_for_float = False
    try:
        float(value_as_str)
        test_for_float = True
    except:
        pass

    if "e" in str(value_as_str).lower() and test_for_float:
        return True
    return False

def is_nan(value: Any) -> bool:
    """
    Returns True if 'value' is some form of NaN, whether float('nan')
    or a numpy or pandas Nan.
    """
    # Test for numpy.nan and float('nan')
    if not value == value:
        return True
    else:
        return False
    


#   Copyright 2020 Connor Ferster

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pathlib
import json
import re
import sys
from types import ModuleType
from forallpeople.dimensions import Dimensions


class Environment:
    """
    A class that contains information about the units definitions that will be used
    by each Physical instance. Each Physical instance requests units definition
    information from the single Environment instance (OMG! Singleton!)
    """

    environment = {}

    def __init__(
        self, physical_class: type, builtins_module: ModuleType, si_base_units: dict
    ):
        self._units_by_dimension = {"derived": dict(), "defined": dict()}
        self._units_by_factor = dict()
        self._physical_class = physical_class
        self._builtins_module = builtins_module
        self._si_base_units = si_base_units
        self.this_module = sys.modules["forallpeople"]
        self.push_module = None
        if not self.environment:
            self.environment = self._si_base_units

    @property
    def units_by_dimension(self):
        def return_dict():
            return self._units_by_dimension
        return return_dict

    @property
    def units_by_factor(self):
        def return_dict():
            return self._units_by_factor
        return return_dict

    def __call__(self, env_name: str = "", top_level: bool = False):
        if not env_name:
            try:
                print(
                        self._generate_units_dict(
                            self.environment, self._physical_class
                        ), "\n", self._si_base_units),
                    
            except TypeError:
                print(self.environment)
            return

        push_module = self.this_module
        if top_level:
            push_module = self._builtins_module

        if self.environment != self._si_base_units and self.push_module:
            old_units_dict = self._generate_units_dict(
                self.environment, self._physical_class
            )

            self.del_vars(old_units_dict, self.push_module)

        self.environment = self._load_environment(env_name)
        new_units_dict = self._generate_units_dict(
            self.environment, self._physical_class
        )
        self.push_vars(new_units_dict, push_module)
        self.push_vars(self._si_base_units, push_module)

        # Update internal class dictionaries: self.units_by_dimension, self.units_by_factor
        self._units_by_dimension = {"derived": dict(), "defined": dict()}
        self._units_by_factor = dict()
        for name, definition in self.environment.items():
            factor = round(
                definition.get("Factor", 1), self._physical_class._total_precision
            )
            dimension = definition.get("Dimension")
            value = definition.get("Value", 1)
            if factor == 1 and value == 1:
                self._units_by_dimension["derived"].setdefault(dimension, dict()).update(
                    {name: definition}
                )
            elif factor != 1:
                self._units_by_dimension["defined"].setdefault(dimension, dict()).update(
                    {name: definition}
                )
                self._units_by_factor.update({factor: {name: definition}})
        self.push_module = push_module  # Update previous push_module; could be either module or top-level

    
    def push_vars(self, units_dict: dict, module: ModuleType) -> None:
        module.__dict__.update(units_dict)

    def del_vars(self, units_dict: dict, module: ModuleType) -> None:
        for key in units_dict.keys():
            module.__dict__.pop(key)

    def _load_environment(self, env_name: str):
        """
        Returns a dict that describes a set of unit definitions as contained in the
        JSON file titled "'env_name'.json". Alternatively, 'env_name' can
        also be a path to a JSON file outside forallpeople.
        After the 'Dimension' definition is converted to an Dimensions
        object and any factors are checked for safety then evaluated.
        Raises error if file not found.
        """
        dim_array_not_defn = (
            "Dimension array not defined in environment"
            " .json file, '{env_name}.json', for unit '{unit}'"
        )
        unit_factor_not_eval = (
            "Unit definition in '{env_name}.json': Factor"
            "must be an arithmetic expr (as a str), a float,"
            "or an int: not '{factor}'."
        )
        if pathlib.Path(env_name).exists():
            file_path = pathlib.Path(env_name)
        else:
            path = pathlib.Path(__file__).parent
            filename = env_name + ".json"
            file_path = path / filename

        with open(file_path, "r", encoding="utf-8") as json_unit_definitions:
            units_environment = json.load(json_unit_definitions)

        # Load definitions
        arithmetic_expr = re.compile(r"[0-9.*/+-]")
        for unit, definitions in units_environment.items():
            dimensions = definitions.get("Dimension", ())
            factor = definitions.get("Factor", "1")
            symbol = definitions.get("Symbol", "")
            if not dimensions:
                raise DimensionError(dim_array_not_defn.format(env_name, unit))
            else:
                units_environment[unit]["Dimension"] = Dimensions(*dimensions)

            if type(factor) is str and not arithmetic_expr.match(factor):
                raise ValueError(unit_factor_not_eval.format(unit, env_name, factor))
            else:
                factor = str(factor)
                units_environment[unit]["Factor"] = eval(factor)
        return units_environment

    @staticmethod
    def _generate_units_dict(environment: dict, physical_class):
        """
        Returns None; updates the globals dict with the units defined in the "definitions"
        portion of the environment dict. This is the method that instantiates all of the
        unit symbols defined in the environment json file.
        """
        units_dict = {}
        # Transfer definitions
        for unit, definitions in environment.items():
            dimensions = definitions["Dimension"]
            factor = definitions.get("Factor", 1)
            symbol = definitions.get("Symbol", "")
            value = definitions.get("Value", 1)
            if symbol:
                units_dict.update(
                    {unit: physical_class(1 / factor, dimensions, factor)}
                )
            else:
                units_dict.update({unit: physical_class(value, dimensions, factor)})
        return units_dict


